from rasa.utils.redactor import Redactor
from rasa.utils.redactor import get_count_digits

redactor = Redactor()
# It would not be wise to try and change these messages unless you have to.
msg_with_pci = """Received user message:
{
   "intent":{
      "name":"features_data_add",
      "confidence":0.10683724284172058
   },
   "entities":[
      {
         "start":0,
         "end":16,
         "text":"4761739001010010",
         "value":"4761739001010010",
         "confidence":1.0,
         "additional_info":{
            "value":4761739001010010,
            "type":"value"
         },
         "entity":"number",
         "extractor":"DucklingHTTPExtractor"
      }
   ],
   "intent_ranking":[
      {
         "name":"features_data_add",
         "confidence":0.10683724284172058
      },
      {
         "name":"this_phone",
         "confidence":0.07678905129432678
      },
      {
         "name":"affirm",
         "confidence":0.07138489931821823
      },
      {
         "name":"deny",
         "confidence":0.06736931204795837
      },
      {
         "name":"tech_support",
         "confidence":0.05541007220745087
      },
      {
         "name":"greet",
         "confidence":0.0402037613093853
      },
      {
         "name":"features_data_inquire",
         "confidence":0.02958683669567108
      },
      {
         "name":"features_data",
         "confidence":0.0289413183927536
      },
      {
         "name":"another_phone",
         "confidence":0.02866683341562748
      },
      {
         "name":"phone_number_change",
         "confidence":0.027637505903840065
      }
   ],
   "text":"4761739001010010",
   "asrConfidence":1,
   "recordingId":"None",
   "missed_intent":false,
   "message_id":"7f3ab5c8e9aa407dbba8553a3ac83671",
   "metadata":{
      "message_type":"DTMF",
      "dnis":"9282252208",
      "privacy_mode":true,
      "channel_id":1,
      "original_text":"4761739001010010",
      "type":"text",
      "active_form":"action_credit_card_collector",
      "requested_slot":"cc_number",
      "template_name":"utter_ask_cc_number",
      "asr_threshold":0.4,
      "nlu_threshold":0.63,
      "group":"cc_number"
   }
}
"""
msg_without_pci = """Received user message:
{
   "intent":{
      "name":"features_data_add",
      "confidence":0.****************
   },
   "entities":[
      {
         "start":0,
         "end":16,
         "text":"****************",
         "value":"****************",
         "confidence":1.0,
         "additional_info":{
            "value":****************,
            "type":"value"
         },
         "entity":"number",
         "extractor":"DucklingHTTPExtractor"
      }
   ],
   "intent_ranking":[
      {
         "name":"features_data_add",
         "confidence":0.****************
      },
      {
         "name":"this_phone",
         "confidence":0.****************
      },
      {
         "name":"affirm",
         "confidence":0.****************
      },
      {
         "name":"deny",
         "confidence":0.****************
      },
      {
         "name":"tech_support",
         "confidence":0.****************
      },
      {
         "name":"greet",
         "confidence":0.****************
      },
      {
         "name":"features_data_inquire",
         "confidence":0.****************
      },
      {
         "name":"features_data",
         "confidence":0.****************
      },
      {
         "name":"another_phone",
         "confidence":0.****************
      },
      {
         "name":"phone_number_change",
         "confidence":0.****************
      }
   ],
   "text":"****************",
   "asrConfidence":1,
   "recordingId":"None",
   "missed_intent":false,
   "message_id":"7f3ab5c8e9aa407dbba8553a3ac83671",
   "metadata":{
      "message_type":"DTMF",
      "dnis":"9282252208",
      "privacy_mode":true,
      "channel_id":1,
      "original_text":"****************",
      "type":"text",
      "active_form":"action_credit_card_collector",
      "requested_slot":"cc_number",
      "template_name":"utter_ask_cc_number",
      "asr_threshold":0.4,
      "nlu_threshold":0.63,
      "group":"cc_number"
   }
}
"""
msg_without_pci_strict = """Received user message:
{
   "intent":{
      "name":"features_data_add",
      "confidence":0.****************
   },
   "entities":[
      {
         "start":0,
         "end":16,
         "text":"****************",
         "value":"****************",
         "confidence":1.0,
         "additional_info":{
            "value":****************,
            "type":"value"
         },
         "entity":"number",
         "extractor":"DucklingHTTPExtractor"
      }
   ],
   "intent_ranking":[
      {
         "name":"features_data_add",
         "confidence":0.****************
      },
      {
         "name":"this_phone",
         "confidence":0.****************
      },
      {
         "name":"affirm",
         "confidence":0.****************
      },
      {
         "name":"deny",
         "confidence":0.****************
      },
      {
         "name":"tech_support",
         "confidence":0.****************
      },
      {
         "name":"greet",
         "confidence":0.****************
      },
      {
         "name":"features_data_inquire",
         "confidence":0.****************
      },
      {
         "name":"features_data",
         "confidence":0.****************
      },
      {
         "name":"another_phone",
         "confidence":0.****************
      },
      {
         "name":"phone_number_change",
         "confidence":0.****************
      }
   ],
   "text":"****************",
   "asrConfidence":1,
   "recordingId":"None",
   "missed_intent":false,
   "message_id":"7f3ab5c8e9aa****************dbba****************a3ac83671",
   "metadata":{
      "message_type":"DTMF",
      "dnis":"9282252208",
      "privacy_mode":true,
      "channel_id":1,
      "original_text":"****************",
      "type":"text",
      "active_form":"action_credit_card_collector",
      "requested_slot":"cc_number",
      "template_name":"utter_ask_cc_number",
      "asr_threshold":0.4,
      "nlu_threshold":0.63,
      "group":"cc_number"
   }
}
"""

dict_with_pci = {
    "intent": {
        "name": "features_data_add",
        "confidence": 0.10683724284172058
    },
    "entities": [
        {
            "start": 0,
            "end": 16,
            "text": "4761739001010010",
            "value": "4761739001010010",
            "confidence": 1.0,
            "additional_info": {
                "value": 4761739001010010,
                "type": "value"
            },
            "entity": "number",
            "extractor": "DucklingHTTPExtractor"
        }
    ],
    "intent_ranking": [
        {
            "name": "features_data_add",
            "confidence": 0.10683724284172058
        },
        {
            "name": "this_phone",
            "confidence": 0.07678905129432678
        },
        {
            "name": "affirm",
            "confidence": 0.07138489931821823
        },
        {
            "name": "deny",
            "confidence": 0.06736931204795837
        },
        {
            "name": "tech_support",
            "confidence": 0.05541007220745087
        },
        {
            "name": "greet",
            "confidence": 0.0402037613093853
        },
        {
            "name": "features_data_inquire",
            "confidence": 0.02958683669567108
        },
        {
            "name": "features_data",
            "confidence": 0.0289413183927536
        },
        {
            "name": "another_phone",
            "confidence": 0.02866683341562748
        },
        {
            "name": "phone_number_change",
            "confidence": 0.027637505903840065
        }
    ],
    "text": "4761739001010010",
    "asrConfidence": 1,
    "recordingId": "None",
    "missed_intent": False,
    "message_id": "7f3ab5c8e9aa407dbba8553a3ac83671",
    "metadata": {
        "message_type": "DTMF",
        "dnis": "9282252208",
        "privacy_mode": True,
        "channel_id": 1,
        "original_text": "4761739001010010",
        "type": "text",
        "active_form": "action_credit_card_collector",
        "requested_slot": "cc_number",
        "template_name": "utter_ask_cc_number",
        "asr_threshold": 0.4,
        "nlu_threshold": 0.63,
        "group": "cc_number"
    },
    "useless_field": None
}
dict_without_pci = {
    "intent": {
        "name": "features_data_add",
        "confidence": 0.10683724284172058
    },
    "entities": [
        {
            "start": 0,
            "end": 16,
            "text": "****************",
            "value": "****************",
            "confidence": 1.0,
            "additional_info": {
                "value": "****************",
                "type": "value"
            },
            "entity": "number",
            "extractor": "DucklingHTTPExtractor"
        }
    ],
    "intent_ranking": [
        {
            "name": "features_data_add",
            "confidence": 0.10683724284172058
        },
        {
            "name": "this_phone",
            "confidence": 0.07678905129432678
        },
        {
            "name": "affirm",
            "confidence": 0.07138489931821823
        },
        {
            "name": "deny",
            "confidence": 0.06736931204795837
        },
        {
            "name": "tech_support",
            "confidence": 0.05541007220745087
        },
        {
            "name": "greet",
            "confidence": 0.0402037613093853
        },
        {
            "name": "features_data_inquire",
            "confidence": 0.02958683669567108
        },
        {
            "name": "features_data",
            "confidence": 0.0289413183927536
        },
        {
            "name": "another_phone",
            "confidence": 0.02866683341562748
        },
        {
            "name": "phone_number_change",
            "confidence": 0.027637505903840065
        }
    ],
    "text": "****************",
    "asrConfidence": 1,
    "recordingId": "None",
    "missed_intent": False,
    "message_id": "7f3ab5c8e9aa407dbba8553a3ac83671",
    "metadata": {
        "message_type": "DTMF",
        "dnis": "9282252208",
        "privacy_mode": True,
        "channel_id": 1,
        "original_text": "****************",
        "type": "text",
        "active_form": "action_credit_card_collector",
        "requested_slot": "cc_number",
        "template_name": "utter_ask_cc_number",
        "asr_threshold": 0.4,
        "nlu_threshold": 0.63,
        "group": "cc_number"
    },
    "useless_field": None
}

list_with_pci = ["4761739001010010", 4761739001010010,
                 {"value": 4761739001010010, "additional_info": {"value": "4761739001010010"}}]
list_without_pci = ["****************", "****************",
                    {"value": "****************", "additional_info": {"value": "****************"}}]
entity_list_one_entry_with_pci = [
    {
        "start": 0,
        "end": 16,
        "text": "4761739001010010",
        "value": "4761739001010010",
        "confidence": 1.0,
        "additional_info": {
            "value": 4761739001010010,
            "type": "value"
        },
        "entity": "number",
        "extractor": "DucklingHTTPExtractor"
    }
]
entity_list_one_entry_without_pci = [
    {
        "start": 0,
        "end": 16,
        "text": "****************",
        "value": "****************",
        "confidence": 1.0,
        "additional_info": {
            "value": "****************",
            "type": "value"
        },
        "entity": "number",
        "extractor": "DucklingHTTPExtractor"
    }
]

entity_list_many_entries_with_pci = [
    {'start': 0, 'end': 1, 'text': '1', 'value': 1, 'confidence': 1.0, 'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 1, 'end': 2, 'text': '1', 'value': 1, 'confidence': 1.0, 'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 2, 'end': 3, 'text': '1', 'value': 1, 'confidence': 1.0, 'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 3, 'end': 4, 'text': '1', 'value': 1, 'confidence': 1.0, 'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 0, 'end': 1, 'text': '1', 'value': 1, 'confidence': 1.0, 'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 1, 'end': 2, 'text': '1', 'value': 1, 'confidence': 1.0, 'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 2, 'end': 3, 'text': '1', 'value': 1, 'confidence': 1.0, 'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 3, 'end': 4, 'text': '1', 'value': 1, 'confidence': 1.0, 'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 4, 'end': 5, 'text': '1', 'value': 1, 'confidence': 1.0, 'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 5, 'end': 6, 'text': '1', 'value': 1, 'confidence': 1.0, 'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 6, 'end': 7, 'text': '1', 'value': 1, 'confidence': 1.0, 'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 7, 'end': 8, 'text': '1', 'value': 1, 'confidence': 1.0, 'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 8, 'end': 9, 'text': '1', 'value': 1, 'confidence': 1.0, 'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 9, 'end': 10, 'text': '1', 'value': 1, 'confidence': 1.0,
     'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 10, 'end': 11, 'text': '1', 'value': 1, 'confidence': 1.0,
     'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
    {'start': 11, 'end': 12, 'text': '1', 'value': 1, 'confidence': 1.0,
     'additional_info': {'value': 1, 'type': 'value'},
     'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
]
entity_list_many_entries_without_pci = [
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 1,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 0,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 2,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 1,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 3,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 2,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 4,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 3,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 1,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 0,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 2,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 1,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 3,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 2,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 4,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 3,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 5,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 4,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 6,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 5,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 7,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 6,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 8,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 7,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 9,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 8,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 10,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 9,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 11,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 10,
     'text': '*',
     'value': '*'},
    {'additional_info': {'type': 'value', 'value': '*'},
     'confidence': 1.0,
     'end': 12,
     'entity': 'number',
     'extractor': 'DucklingHTTPExtractor',
     'start': 11,
     'text': '*',
     'value': '*'}]

event_no_pci = {'event': 'user', 'timestamp': 1621281543.1603343,
                'metadata': {'dnis': '9282252467', 'message_type': 'SPOKEN',
                             'partials': ['Ok', 'Ok so', 'Ok so', "Ok so i've been", "Ok so i've been having",
                                          "Ok so i've been having an effect", "Ok so i've been having an issue",
                                          "Ok so i've been having an issue with",
                                          "Ok so i've been having an issue where",
                                          "Ok so i've been having an issue where some",
                                          "Ok so i've been having an issue where some static",
                                          "Ok so i've been having an issue where some static i want to",
                                          "Ok so i've been having an issue where some static i want to cross the line",
                                          "Ok so i've been having an issue where some static i want to cross the line on the phone",
                                          "Ok so i've been having an issue where some static i want to cross the line on the phone",
                                          "Ok so i've been having an issue where some static i want to cross the line on the phone and"],
                             'channel_id': 1, 'type': 'text', 'template_name': 'temp_help_message',
                             'asr_threshold': 0.2, 'nlu_threshold': 0.63},
                'text': "OK, so I've been having an issue with some static, I want to cross the line on my phone and it",
                'parse_data': {'intent': {'name': 'add_subscriber', 'confidence': 0.6653156280517578}, 'entities': [],
                               'intent_ranking': [{'name': 'add_subscriber', 'confidence': 0.6653156280517578},
                                                  {'name': 'plan_info', 'confidence': 0.06193934753537178},
                                                  {'name': 'phone_number_change', 'confidence': 0.053489990532398224},
                                                  {'name': 'autopay', 'confidence': 0.013421786949038506},
                                                  {'name': 'autopay_update', 'confidence': 0.013023709878325462},
                                                  {'name': 'pay_someone_else', 'confidence': 0.00991435069590807},
                                                  {'name': 'remove_subscriber', 'confidence': 0.009756733663380146},
                                                  {'name': 'reactivate', 'confidence': 0.009636330418288708},
                                                  {'name': 'change_plan', 'confidence': 0.009616482071578503},
                                                  {'name': 'another_phone', 'confidence': 0.009423586539924145}],
                               'text': "Ok so i've been having an issue where some static i want to cross the line on the phone",
                               'asrConfidence': 1, 'recordingId': None, 'missed_intent': None,
                               'message_id': '6524399f26124e8d87957f6c6ecb5a39',
                               'metadata': {'dnis': '9282252467', 'message_type': 'SPOKEN',
                                            'partials': ['Ok', 'Ok so', 'Ok so', "Ok so i've been",
                                                         "Ok so i've been having", "Ok so i've been having an effect",
                                                         "Ok so i've been having an issue",
                                                         "Ok so i've been having an issue with",
                                                         "Ok so i've been having an issue where",
                                                         "Ok so i've been having an issue where some",
                                                         "Ok so i've been having an issue where some static",
                                                         "Ok so i've been having an issue where some static i want to",
                                                         "Ok so i've been having an issue where some static i want to cross the line",
                                                         "Ok so i've been having an issue where some static i want to cross the line on the phone",
                                                         "Ok so i've been having an issue where some static i want to cross the line on the phone",
                                                         "Ok so i've been having an issue where some static i want to cross the line on the phone and"],
                                            'channel_id': 1, 'type': 'text', 'template_name': 'temp_help_message',
                                            'asr_threshold': 0.2, 'nlu_threshold': 0.63}},
                'input_channel': 'OMNI_VOICE_00001', 'message_id': '6524399f26124e8d87957f6c6ecb5a39',
                'entities_redacted': []}
event_with_pci = {'event': 'user', 'timestamp': 1621281543.1603343,
                  'metadata': {'dnis': '9282252467', 'message_type': 'SPOKEN',
                               'partials': ['1', '1 2', '1 2', "1 2 3"],
                               'channel_id': 1, 'type': 'text', 'template_name': 'temp_help_message',
                               'asr_threshold': 0.2, 'nlu_threshold': 0.63},
                  'text': "1 2 3",
                  'parse_data': {'intent': {'name': 'add_subscriber', 'confidence': 0.6653156280517578},
                                 'entities': [{'start': 0, 'end': 1, 'text': '1', 'value': 1, 'confidence': 1.0,
                                               'additional_info': {'value': 1, 'type': 'value'},
                                               'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                              {'start': 1, 'end': 2, 'text': '2', 'value': 2, 'confidence': 1.0,
                                               'additional_info': {'value': 2, 'type': 'value'},
                                               'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                              {'start': 2, 'end': 3, 'text': '3', 'value': 3, 'confidence': 1.0,
                                               'additional_info': {'value': 3, 'type': 'value'},
                                               'entity': 'number', 'extractor': 'DucklingHTTPExtractor'}],
                                 'intent_ranking': [{'name': 'add_subscriber', 'confidence': 0.6653156280517578},
                                                    {'name': 'plan_info', 'confidence': 0.06193934753537178},
                                                    {'name': 'phone_number_change',
                                                     'confidence': 0.053489990532398224},
                                                    {'name': 'autopay', 'confidence': 0.013421786949038506},
                                                    {'name': 'autopay_update', 'confidence': 0.013023709878325462},
                                                    {'name': 'pay_someone_else', 'confidence': 0.00991435069590807},
                                                    {'name': 'remove_subscriber', 'confidence': 0.009756733663380146},
                                                    {'name': 'reactivate', 'confidence': 0.009636330418288708},
                                                    {'name': 'change_plan', 'confidence': 0.009616482071578503},
                                                    {'name': 'another_phone', 'confidence': 0.009423586539924145}],
                                 'text': "1 2 3",
                                 'asrConfidence': 1, 'recordingId': None, 'missed_intent': None,
                                 'message_id': '6524399f26124e8d87957f6c6ecb5a39',
                                 'metadata': {'dnis': '9282252467', 'message_type': 'SPOKEN',
                                              'partials': ['1', '1 2', '1 2', "1 2 3"],
                                              'channel_id': 1, 'type': 'text', 'template_name': 'temp_help_message',
                                              'asr_threshold': 0.2, 'nlu_threshold': 0.63}},
                  'input_channel': 'OMNI_VOICE_00001', 'message_id': '6524399f26124e8d87957f6c6ecb5a39',
                  'entities_redacted': [{'start': 0, 'end': 1, 'text': '1', 'value': 1, 'confidence': 1.0,
                                         'additional_info': {'value': 1, 'type': 'value'},
                                         'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                        {'start': 1, 'end': 2, 'text': '2', 'value': 2, 'confidence': 1.0,
                                         'additional_info': {'value': 2, 'type': 'value'},
                                         'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                        {'start': 2, 'end': 3, 'text': '3', 'value': 3, 'confidence': 1.0,
                                         'additional_info': {'value': 3, 'type': 'value'},
                                         'entity': 'number', 'extractor': 'DucklingHTTPExtractor'}]}
event_with_redacted_pci = {'event': 'user', 'timestamp': 1621281543.1603343,
                           'metadata': {'dnis': '9282252467', 'message_type': 'SPOKEN',
                                        'partials': ['****', '**** ****', '**** ****', "**** **** ****"],
                                        'channel_id': 1, 'type': 'text', 'template_name': 'temp_help_message',
                                        'asr_threshold': 0.2, 'nlu_threshold': 0.63},
                           'text': "**** **** ****",
                           'parse_data': {'intent': {'name': 'add_subscriber', 'confidence': 0.6653156280517578},
                                          'entities': [
                                              {'start': 0, 'end': 1, 'text': '*', 'value': '*', 'confidence': 1.0,
                                               'additional_info': {'value': '*', 'type': 'value'},
                                               'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                              {'start': 1, 'end': 2, 'text': '*', 'value': '*', 'confidence': 1.0,
                                               'additional_info': {'value': '*', 'type': 'value'},
                                               'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                              {'start': 2, 'end': 3, 'text': '*', 'value': '*', 'confidence': 1.0,
                                               'additional_info': {'value': '*', 'type': 'value'},
                                               'entity': 'number', 'extractor': 'DucklingHTTPExtractor'}],
                                          'intent_ranking': [
                                              {'name': 'add_subscriber', 'confidence': 0.6653156280517578},
                                              {'name': 'plan_info', 'confidence': 0.06193934753537178},
                                              {'name': 'phone_number_change',
                                               'confidence': 0.053489990532398224},
                                              {'name': 'autopay', 'confidence': 0.013421786949038506},
                                              {'name': 'autopay_update', 'confidence': 0.013023709878325462},
                                              {'name': 'pay_someone_else', 'confidence': 0.00991435069590807},
                                              {'name': 'remove_subscriber', 'confidence': 0.009756733663380146},
                                              {'name': 'reactivate', 'confidence': 0.009636330418288708},
                                              {'name': 'change_plan', 'confidence': 0.009616482071578503},
                                              {'name': 'another_phone', 'confidence': 0.009423586539924145}],
                                          'text': "**** **** ****",
                                          'asrConfidence': 1, 'recordingId': None, 'missed_intent': None,
                                          'message_id': '6524399f26124e8d87957f6c6ecb5a39',
                                          'metadata': {'dnis': '9282252467', 'message_type': 'SPOKEN',
                                                       'partials': ['****', '**** ****', '**** ****', "**** **** ****"],
                                                       'channel_id': 1, 'type': 'text',
                                                       'template_name': 'temp_help_message',
                                                       'asr_threshold': 0.2, 'nlu_threshold': 0.63}},
                           'input_channel': 'OMNI_VOICE_00001', 'message_id': '6524399f26124e8d87957f6c6ecb5a39',
                           'entities_redacted': [{'start': 0, 'end': 1, 'text': '*', 'value': '*', 'confidence': 1.0,
                                                  'additional_info': {'value': '*', 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 1, 'end': 2, 'text': '*', 'value': '*', 'confidence': 1.0,
                                                  'additional_info': {'value': '*', 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 2, 'end': 3, 'text': '*', 'value': '*', 'confidence': 1.0,
                                                  'additional_info': {'value': '*', 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'}]}
event_with_cc_num = {'event': 'user', 'timestamp': 1621281543.1603343,
                     'metadata': {'dnis': '9282252467', 'message_type': 'SPOKEN',
                                  'partials': ['1', '1 2', '1 2', "1 2 3"],
                                  'channel_id': 1, 'type': 'text', 'template_name': 'temp_help_message',
                                  'asr_threshold': 0.2, 'nlu_threshold': 0.63},
                     'text': "1111111111111",
                     'parse_data': {'intent': {'name': 'add_subscriber', 'confidence': 0.6653156280517578},
                                    'entities': [{'start': 0, 'end': 1, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 1, 'end': 2, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 2, 'end': 3, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 3, 'end': 4, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 0, 'end': 1, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 1, 'end': 2, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 2, 'end': 3, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 3, 'end': 4, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 4, 'end': 5, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 5, 'end': 6, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 6, 'end': 7, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 7, 'end': 8, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 8, 'end': 9, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 9, 'end': 10, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 10, 'end': 11, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                                                 {'start': 11, 'end': 12, 'text': '1', 'value': 1, 'confidence': 1.0,
                                                  'additional_info': {'value': 1, 'type': 'value'},
                                                  'entity': 'number', 'extractor': 'DucklingHTTPExtractor'}],
                                    'intent_ranking': [{'name': 'add_subscriber', 'confidence': 0.6653156280517578},
                                                       {'name': 'plan_info', 'confidence': 0.06193934753537178},
                                                       {'name': 'phone_number_change',
                                                        'confidence': 0.053489990532398224},
                                                       {'name': 'autopay', 'confidence': 0.013421786949038506},
                                                       {'name': 'autopay_update', 'confidence': 0.013023709878325462},
                                                       {'name': 'pay_someone_else', 'confidence': 0.00991435069590807},
                                                       {'name': 'remove_subscriber',
                                                        'confidence': 0.009756733663380146},
                                                       {'name': 'reactivate', 'confidence': 0.009636330418288708},
                                                       {'name': 'change_plan', 'confidence': 0.009616482071578503},
                                                       {'name': 'another_phone', 'confidence': 0.009423586539924145}],
                                    'text': "1111111111111",
                                    'asrConfidence': 1, 'recordingId': None, 'missed_intent': None,
                                    'message_id': '6524399f26124e8d87957f6c6ecb5a39',
                                    'metadata': {'dnis': '9282252467', 'message_type': 'SPOKEN',
                                                 'partials': ['1', '1 2', '1 2', "1 2 3"],
                                                 'channel_id': 1, 'type': 'text', 'template_name': 'temp_help_message',
                                                 'asr_threshold': 0.2, 'nlu_threshold': 0.63}},
                     'input_channel': 'OMNI_VOICE_00001', 'message_id': '6524399f26124e8d87957f6c6ecb5a39',
                     'entities_redacted': [
                         {'start': 0, 'end': 1, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 1, 'end': 2, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 2, 'end': 3, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 3, 'end': 4, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 0, 'end': 1, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 1, 'end': 2, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 2, 'end': 3, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 3, 'end': 4, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 4, 'end': 5, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 5, 'end': 6, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 6, 'end': 7, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 7, 'end': 8, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 8, 'end': 9, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 9, 'end': 10, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 10, 'end': 11, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'},
                         {'start': 11, 'end': 12, 'text': '1', 'value': 1, 'confidence': 1.0,
                          'additional_info': {'value': 1, 'type': 'value'},
                          'entity': 'number', 'extractor': 'DucklingHTTPExtractor'}]}
event_with_cc_num_redacted = {
    'entities_redacted': [
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 1,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 0,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 2,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 1,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 3,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 2,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 4,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 3,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 1,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 0,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 2,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 1,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 3,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 2,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 4,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 3,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 5,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 4,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 6,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 5,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 7,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 6,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 8,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 7,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 9,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 8,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 10,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 9,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 11,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 10,
         'text': '*',
         'value': '*'},
        {'additional_info': {'type': 'value', 'value': '*'},
         'confidence': 1.0,
         'end': 12,
         'entity': 'number',
         'extractor': 'DucklingHTTPExtractor',
         'start': 11,
         'text': '*',
         'value': '*'}],
    'event': 'user',
    'input_channel': 'OMNI_VOICE_00001',
    'message_id': '6524399f26124e8d87957f6c6ecb5a39',
    'metadata': {'asr_threshold': 0.2,
                 'channel_id': 1,
                 'dnis': '9282252467',
                 'message_type': 'SPOKEN',
                 'nlu_threshold': 0.63,
                 'partials': ['1', '1 2', '1 2', '1 2 3'],
                 'template_name': 'temp_help_message',
                 'type': 'text'},
    'parse_data': {'asrConfidence': 1,
                   'entities': [
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 1,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 0,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 2,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 1,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 3,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 2,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 4,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 3,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 1,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 0,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 2,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 1,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 3,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 2,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 4,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 3,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 5,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 4,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 6,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 5,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 7,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 6,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 8,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 7,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 9,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 8,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 10,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 9,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 11,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 10,
                        'text': '*',
                        'value': '*'},
                       {'additional_info': {'type': 'value',
                                            'value': '*'},
                        'confidence': 1.0,
                        'end': 12,
                        'entity': 'number',
                        'extractor': 'DucklingHTTPExtractor',
                        'start': 11,
                        'text': '*',
                        'value': '*'}],
                   'intent': {'confidence': 0.6653156280517578,
                              'name': 'add_subscriber'},
                   'intent_ranking': [
                       {'confidence': 0.6653156280517578,
                        'name': 'add_subscriber'},
                       {'confidence': 0.06193934753537178,
                        'name': 'plan_info'},
                       {'confidence': 0.053489990532398224,
                        'name': 'phone_number_change'},
                       {'confidence': 0.013421786949038506,
                        'name': 'autopay'},
                       {'confidence': 0.013023709878325462,
                        'name': 'autopay_update'},
                       {'confidence': 0.00991435069590807,
                        'name': 'pay_someone_else'},
                       {'confidence': 0.009756733663380146,
                        'name': 'remove_subscriber'},
                       {'confidence': 0.009636330418288708,
                        'name': 'reactivate'},
                       {'confidence': 0.009616482071578503,
                        'name': 'change_plan'},
                       {'confidence': 0.009423586539924145,
                        'name': 'another_phone'}],
                   'message_id': '6524399f26124e8d87957f6c6ecb5a39',
                   'metadata': {'asr_threshold': 0.2,
                                'channel_id': 1,
                                'dnis': '9282252467',
                                'message_type': 'SPOKEN',
                                'nlu_threshold': 0.63,
                                'partials': ['1', '1 2', '1 2', '1 2 3'],
                                'template_name': 'temp_help_message',
                                'type': 'text'},
                   'missed_intent': None,
                   'recordingId': None,
                   'text': '****************'},
    'text': '****************',
    'timestamp': 1621281543.1603343}

event_with_cc_num_2 = {
    'sender_id': '05fa294af33745a5b34ee30c7b06b996',
    'event': 'user',
    'timestamp': 1627090353.228758,
    'metadata': {
        'replyType': 'text',
        'channel_id': 1,
        'original_text': 'my cc is 1111111111111111',
        'should_redact': False,
        'type': 'text',
        'template_name': 'utter_greet',
        'asr_threshold': 0.2,
        'nlu_threshold': 0.67,
        'group': 'SALUTATION'
    },
    'text': 'my cc is 1111111111111111',
    'parse_data': {
        'intent': {
            'name': 'None',
            'confidence': 0.5955944657325745
        },
        'entities': [{
            'start': 9,
            'end': 12,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 13,
            'end': 16,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 17,
            'end': 20,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 21,
            'end': 24,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 25,
            'end': 28,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 29,
            'end': 32,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 33,
            'end': 36,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 37,
            'end': 40,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 41,
            'end': 44,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 45,
            'end': 48,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 49,
            'end': 52,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 53,
            'end': 56,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 57,
            'end': 60,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 61,
            'end': 64,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 65,
            'end': 68,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 69,
            'end': 72,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }],
        'intent_ranking': [{
            'name': 'None',
            'confidence': 0.5955944657325745
        }, {
            'name': 'card',
            'confidence': 0.4034712314605713
        }, {
            'name': 'freeze',
            'confidence': 0.0006689034635201097
        }, {
            'name': 'language',
            'confidence': 0.00012367829913273454
        }, {
            'name': 'autopay',
            'confidence': 0.00010715128155425191
        }, {
            'name': 'disconnect',
            'confidence': 1.715622602205258e-05
        }, {
            'name': 'handoff',
            'confidence': 9.751319339557085e-06
        }, {
            'name': 'thanks',
            'confidence': 3.094106205026037e-06
        }, {
            'name': 'transfer',
            'confidence': 3.039700004592305e-06
        }, {
            'name': 'deny',
            'confidence': 1.0337059848097851e-06
        }],
        'text': 'my cc is 1111111111111111',
        'asrConfidence': 1,
        'recordingId': None,
        'missed_intent': True,
        'message_id': '10837e5268ca40579d8fc90ba63a554c',
        'metadata': {
            'replyType': 'text',
            'channel_id': 1,
            'original_text': 'my cc is 1111111111111111',
            'should_redact': False,
            'type': 'text',
            'template_name': 'utter_greet',
            'asr_threshold': 0.2,
            'nlu_threshold': 0.67,
            'group': 'SALUTATION'
        }
    },
    'input_channel': 'OMNI_WEBCHAT_00001',
    'message_id': '10837e5268ca40579d8fc90ba63a554c',
    'entities_redacted': [
        {
            'start': 9,
            'end': 12,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 13,
            'end': 16,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 17,
            'end': 20,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 21,
            'end': 24,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 25,
            'end': 28,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 29,
            'end': 32,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 33,
            'end': 36,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 37,
            'end': 40,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 41,
            'end': 44,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 45,
            'end': 48,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 49,
            'end': 52,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 53,
            'end': 56,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 57,
            'end': 60,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 61,
            'end': 64,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 65,
            'end': 68,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 69,
            'end': 72,
            'text': '1',
            'value': 1,
            'confidence': 1.0,
            'additional_info': {
                'value': 1,
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }]
}
event_with_cc_num_2_redacted = {
    'sender_id': '05fa294af33745a5b34ee30c7b06b996',
    'event': 'user',
    'timestamp': 1627090353.228758,
    'metadata': {
        'replyType': 'text',
        'channel_id': 1,
        'original_text': 'my cc is ****************',
        'should_redact': False,
        'type': 'text',
        'template_name': 'utter_greet',
        'asr_threshold': 0.2,
        'nlu_threshold': 0.67,
        'group': 'SALUTATION'
    },
    'text': 'my cc is ****************',
    'parse_data': {
        'intent': {
            'name': 'None',
            'confidence': 0.5955944657325745
        },
        'entities': [
            {
                'start': 9,
                'end': 12,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 13,
                'end': 16,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 17,
                'end': 20,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 21,
                'end': 24,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 25,
                'end': 28,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 29,
                'end': 32,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 33,
                'end': 36,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 37,
                'end': 40,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 41,
                'end': 44,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 45,
                'end': 48,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 49,
                'end': 52,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 53,
                'end': 56,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 57,
                'end': 60,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 61,
                'end': 64,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 65,
                'end': 68,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }, {
                'start': 69,
                'end': 72,
                'text': '*',
                'value': '*',
                'confidence': 1.0,
                'additional_info': {
                    'value': '*',
                    'type': 'value'
                },
                'entity': 'number',
                'extractor': 'DucklingHTTPExtractor'
            }],
        'intent_ranking': [
            {
                'name': 'None',
                'confidence': 0.5955944657325745
            }, {
                'name': 'card',
                'confidence': 0.4034712314605713
            }, {
                'name': 'freeze',
                'confidence': 0.0006689034635201097
            }, {
                'name': 'language',
                'confidence': 0.00012367829913273454
            }, {
                'name': 'autopay',
                'confidence': 0.00010715128155425191
            }, {
                'name': 'disconnect',
                'confidence': 1.715622602205258e-05
            }, {
                'name': 'handoff',
                'confidence': 9.751319339557085e-06
            }, {
                'name': 'thanks',
                'confidence': 3.094106205026037e-06
            }, {
                'name': 'transfer',
                'confidence': 3.039700004592305e-06
            }, {
                'name': 'deny',
                'confidence': 1.0337059848097851e-06
            }],
        'text': 'my cc is ****************',
        'asrConfidence': 1,
        'recordingId': None,
        'missed_intent': True,
        'message_id': '10837e5268ca40579d8fc90ba63a554c',
        'metadata': {
            'replyType': 'text',
            'channel_id': 1,
            'original_text': 'my cc is ****************',
            'should_redact': False,
            'type': 'text',
            'template_name': 'utter_greet',
            'asr_threshold': 0.2,
            'nlu_threshold': 0.67,
            'group': 'SALUTATION'
        }
    },
    'input_channel': 'OMNI_WEBCHAT_00001',
    'message_id': '10837e5268ca40579d8fc90ba63a554c',
    'entities_redacted': [
        {
            'start': 9,
            'end': 12,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 13,
            'end': 16,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 17,
            'end': 20,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 21,
            'end': 24,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 25,
            'end': 28,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 29,
            'end': 32,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 33,
            'end': 36,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 37,
            'end': 40,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 41,
            'end': 44,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 45,
            'end': 48,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 49,
            'end': 52,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 53,
            'end': 56,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 57,
            'end': 60,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 61,
            'end': 64,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 65,
            'end': 68,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }, {
            'start': 69,
            'end': 72,
            'text': '*',
            'value': '*',
            'confidence': 1.0,
            'additional_info': {
                'value': '*',
                'type': 'value'
            },
            'entity': 'number',
            'extractor': 'DucklingHTTPExtractor'
        }]
}


def test_string_redaction():
    assert redactor.redact_string(msg_with_pci) == msg_without_pci

def test_string_redaction_spacing():
    msg = "1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
    msg_redacted = '****************'
    assert redactor.redact_string(msg) == msg_redacted


def test_string_redaction_strict():
    assert redactor.redact_string_strict(msg_with_pci) == msg_without_pci_strict

def test_string_redaction_spacing_strict():
    msg = "1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1"
    msg_redacted = '****************'
    assert redactor.redact_string_strict(msg) == msg_redacted

def test_redact_int_len_1():
    assert redactor.redact_int(1) == 1


def test_redact_int_len_1_strict():
    assert redactor.redact_int_strict(1) == 1


def test_redact_int_len_3_strict():
    assert redactor.redact_int_strict(100) == '***'


def test_redact_int_len_4_strict():
    assert redactor.redact_int_strict(1001) == '****'


def test_redact_int_len_6_strict():
    assert redactor.redact_int_strict(100123) == '******'


def test_redact_int_len_10_strict():
    assert redactor.redact_int(1234567890) == 1234567890


def test_redact_int_len_10():
    assert redactor.redact_int(1234567890) == 1234567890


def test_redact_int_len_11():
    assert redactor.redact_int(12345678901) == 12345678901


def test_redact_int_len_12():
    assert redactor.redact_int(123456789012) == 123456789012


def test_redact_int_len_13():
    assert redactor.redact_int(1234567890123) == '*************'


def test_redact_int_len_15():
    assert redactor.redact_int(123456789012345) == '***************'


def test_redact_int_len_16():
    assert redactor.redact_int(1234567890123456) == '****************'


def test_redact_int_len_19():
    assert redactor.redact_int(1234567890123456789) == '*******************'


def test_redact_float_not_int():
    assert redactor.redact_float(123456789012345.6) == 123456789012345.6


def test_redact_float_low_digit_count_int():
    assert redactor.redact_float(1234567890.0) == 1234567890.0


def test_redact_float_high_digit_count_int():
    assert redactor.redact_float(4761739001010010.0) == '****************'


def test_redact_list():
    assert redactor.redact_list(list_with_pci) == list_without_pci


def test_redact_entity_list_one_entry():
    assert redactor.redact_entity_list(entity_list_one_entry_with_pci) == entity_list_one_entry_without_pci


def test_redact_entity_list_many_entries():
    assert redactor.redact_entity_list(entity_list_many_entries_with_pci) == entity_list_many_entries_without_pci


def test_redact_dictionary():
    assert redactor.redact_dict(dict_with_pci) == dict_without_pci


def test_redact_event_no_pci():
    assert redactor.redact_event(event_no_pci) == event_no_pci


def test_redact_event_full_pci_redact():
    assert redactor.redact_event_complete(event_with_pci) == event_with_redacted_pci


def test_redact_event_general_redaction():
    assert redactor.redact_event(event_with_cc_num) == event_with_cc_num_redacted


def test_redact_event_general_redaction_2():
    assert redactor.redact_event(event_with_cc_num_2) == event_with_cc_num_2_redacted


def test_get_digit_length():
    assert get_count_digits(-99999999999999999999) == 20
    assert get_count_digits(-10000000000000000000) == 20
    assert get_count_digits(-9999999999999999999) == 19
    assert get_count_digits(-1000000000000000000) == 19
    assert get_count_digits(-999999999999999999) == 18
    assert get_count_digits(-100000000000000000) == 18
    assert get_count_digits(-99999999999999999) == 17
    assert get_count_digits(-10000000000000000) == 17
    assert get_count_digits(-9999999999999999) == 16
    assert get_count_digits(-1000000000000000) == 16
    assert get_count_digits(-999999999999999) == 15
    assert get_count_digits(-100000000000000) == 15
    assert get_count_digits(-99999999999999) == 14
    assert get_count_digits(-10000000000000) == 14
    assert get_count_digits(-9999999999999) == 13
    assert get_count_digits(-1000000000000) == 13
    assert get_count_digits(-999999999999) == 12
    assert get_count_digits(-100000000000) == 12
    assert get_count_digits(-99999999999) == 11
    assert get_count_digits(-10000000000) == 11
    assert get_count_digits(-9999999999) == 10
    assert get_count_digits(-1000000000) == 10
    assert get_count_digits(-999999999) == 9
    assert get_count_digits(-100000000) == 9
    assert get_count_digits(-99999999) == 8
    assert get_count_digits(-10000000) == 8
    assert get_count_digits(-9999999) == 7
    assert get_count_digits(-1000000) == 7
    assert get_count_digits(-999999) == 6
    assert get_count_digits(-100000) == 6
    assert get_count_digits(-99999) == 5
    assert get_count_digits(-10000) == 5
    assert get_count_digits(-9999) == 4
    assert get_count_digits(-1000) == 4
    assert get_count_digits(-999) == 3
    assert get_count_digits(-100) == 3
    assert get_count_digits(-99) == 2
    assert get_count_digits(-10) == 2
    assert get_count_digits(-9) == 1
    assert get_count_digits(-1) == 1
    assert get_count_digits(0) == 1
    assert get_count_digits(1) == 1
    assert get_count_digits(9) == 1
    assert get_count_digits(10) == 2
    assert get_count_digits(99) == 2
    assert get_count_digits(100) == 3
    assert get_count_digits(999) == 3
    assert get_count_digits(1000) == 4
    assert get_count_digits(9999) == 4
    assert get_count_digits(10000) == 5
    assert get_count_digits(99999) == 5
    assert get_count_digits(100000) == 6
    assert get_count_digits(999999) == 6
    assert get_count_digits(1000000) == 7
    assert get_count_digits(9999999) == 7
    assert get_count_digits(10000000) == 8
    assert get_count_digits(99999999) == 8
    assert get_count_digits(100000000) == 9
    assert get_count_digits(999999999) == 9
    assert get_count_digits(1000000000) == 10
    assert get_count_digits(9999999999) == 10
    assert get_count_digits(10000000000) == 11
    assert get_count_digits(99999999999) == 11
    assert get_count_digits(100000000000) == 12
    assert get_count_digits(999999999999) == 12
    assert get_count_digits(1000000000000) == 13
    assert get_count_digits(9999999999999) == 13
    assert get_count_digits(10000000000000) == 14
    assert get_count_digits(99999999999999) == 14
    assert get_count_digits(100000000000000) == 15
    assert get_count_digits(999999999999999) == 15
    assert get_count_digits(1000000000000000) == 16
    assert get_count_digits(9999999999999999) == 16
    assert get_count_digits(10000000000000000) == 17
    assert get_count_digits(99999999999999999) == 17
    assert get_count_digits(100000000000000000) == 18
    assert get_count_digits(999999999999999999) == 18
    assert get_count_digits(1000000000000000000) == 19
    assert get_count_digits(9999999999999999999) == 19
    assert get_count_digits(10000000000000000000) == 20
    assert get_count_digits(99999999999999999999) == 20



import logging
import typing
from typing import Any, Dict, Optional, Text

from rasa.nlu.classifiers.embedding_intent_classifier import EmbeddingIntentClassifier
from rasa.core.actions.action import RESPOND_PREFIX
from rasa.nlu.constants import (
    MESSAGE_RESPONSE_ATTRIBUTE,
    MESSAGE_INTENT_ATTRIBUTE,
    MESSAGE_TEXT_ATTRIBUTE,
    MESSAGE_TOKENS_NAMES,
    MESSAGE_ATTRIBUTES,
    MESSAGE_SPACY_FEATURES_NAMES,
    MESSAGE_VECTOR_FEATURE_NAMES,
    OPEN_UTTERANCE_PREDICTION_KEY,
    OPEN_UTTERANCE_RANKING_KEY,
    MESSAGE_SELECTOR_PROPERTY_NAME,
    DEFAULT_OPEN_UTTERANCE_TYPE,
)

logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from rasa.nlu.training_data import Message

import tensorflow as tf

# avoid warning println on contrib import - remove for tf 2
tf.contrib._warning = None


class ResponseSelector(EmbeddingIntentClassifier):
    """Response selector using supervised embeddings.

    The response selector embeds user inputs
    and candidate response into the same space.
    Supervised embeddings are trained by maximizing similarity between them.
    It also provides rankings of the response that did not "win".

    The supervised response selector needs to be preceded by
    a featurizer in the pipeline.
    This featurizer creates the features used for the embeddings.
    It is recommended to use ``CountVectorsFeaturizer`` that
    can be optionally preceded by ``SpacyNLP`` and ``SpacyTokenizer``.

    Based on the starspace idea from: https://arxiv.org/abs/1709.03856.
    However, in this implementation the `mu` parameter is treated differently
    and additional hidden layers are added together with dropout.
    """

    provides = ["response", "response_ranking"]

    requires = [MESSAGE_VECTOR_FEATURE_NAMES[MESSAGE_TEXT_ATTRIBUTE]]

    # default properties (DOC MARKER - don't remove)
    defaults = {
        # nn architecture
        # sizes of hidden layers before the embedding layer for input words
        # the number of hidden layers is thus equal to the length of this list
        "hidden_layers_sizes_a": [256, 128],
        # sizes of hidden layers before the embedding layer for intent labels
        # the number of hidden layers is thus equal to the length of this list
        "hidden_layers_sizes_b": [256, 128],
        # Whether to share the hidden layer weights between input words and intent labels
        "share_hidden_layers": False,
        # training parameters
        # initial and final batch sizes - batch size will be
        # linearly increased for each epoch
        "batch_size": [64, 256],
        # how to create batches
        "batch_strategy": "balanced",  # string 'sequence' or 'balanced'
        # number of epochs
        "epochs": 300,
        # set random seed to any int to get reproducible results
        "random_seed": None,
        # embedding parameters
        # dimension size of embedding vectors
        "embed_dim": 20,
        # the type of the similarity
        "num_neg": 20,
        # flag if minimize only maximum similarity over incorrect actions
        "similarity_type": "auto",  # string 'auto' or 'cosine' or 'inner'
        # the type of the loss function
        "loss_type": "softmax",  # string 'softmax' or 'margin'
        # how similar the algorithm should try
        # to make embedding vectors for correct intent labels
        "mu_pos": 0.8,  # should be 0.0 < ... < 1.0 for 'cosine'
        # maximum negative similarity for incorrect intent labels
        "mu_neg": -0.4,  # should be -1.0 < ... < 1.0 for 'cosine'
        # flag: if true, only minimize the maximum similarity for
        # incorrect intent labels
        "use_max_sim_neg": True,
        # scale loss inverse proportionally to confidence of correct prediction
        "scale_loss": True,
        # regularization parameters
        # the scale of L2 regularization
        "C2": 0.002,
        # the scale of how critical the algorithm should be of minimizing the
        # maximum similarity between embeddings of different intent labels
        "C_emb": 0.8,
        # dropout rate for rnn
        "droprate": 0.2,
        # visualization of accuracy
        # how often to calculate training accuracy
        "evaluate_every_num_epochs": 20,  # small values may hurt performance
        # how many examples to use for calculation of training accuracy
        "evaluate_on_num_examples": 0,  # large values may hurt performance,
        # selector config
        # name of the intent for which this response selector is to be trained
        "retrieval_intent": None,
    }
    # end default properties (DOC MARKER - don't remove)

    def _load_selector_params(self, config: Dict[Text, Any]):
        self.retrieval_intent = config["retrieval_intent"]
        if not self.retrieval_intent:
            # retrieval intent was left to its default value
            logger.info(
                "Retrieval intent parameter was left to its default value. This response selector will be trained"
                "on training examples combining all retrieval intents."
            )

    def _load_params(self) -> None:
        super()._load_params()
        self._load_selector_params(self.component_config)

    @staticmethod
    def _set_message_property(
        message: "Message", prediction_dict: Dict[Text, Any], selector_key: Text
    ):

        message_selector_properties = message.get(MESSAGE_SELECTOR_PROPERTY_NAME, {})
        message_selector_properties[selector_key] = prediction_dict
        message.set(
            MESSAGE_SELECTOR_PROPERTY_NAME,
            message_selector_properties,
            add_to_output=True,
        )

    def preprocess_train_data(self, training_data):
        """Performs sanity checks on training data, extracts encodings for labels and prepares data for training"""

        if self.retrieval_intent:
            training_data = training_data.filter_by_intent(self.retrieval_intent)

        label_id_dict = self._create_label_id_dict(
            training_data, attribute=MESSAGE_RESPONSE_ATTRIBUTE
        )

        self.inverted_label_dict = {v: k for k, v in label_id_dict.items()}
        self._encoded_all_label_ids = self._create_encoded_label_ids(
            training_data,
            label_id_dict,
            attribute=MESSAGE_RESPONSE_ATTRIBUTE,
            attribute_feature_name=MESSAGE_VECTOR_FEATURE_NAMES[
                MESSAGE_RESPONSE_ATTRIBUTE
            ],
        )

        # check if number of negatives is less than number of label_ids
        logger.debug(
            "Check if num_neg {} is smaller than "
            "number of label_ids {}, "
            "else set num_neg to the number of label_ids - 1"
            "".format(self.num_neg, self._encoded_all_label_ids.shape[0])
        )
        # noinspection PyAttributeOutsideInit
        self.num_neg = min(self.num_neg, self._encoded_all_label_ids.shape[0] - 1)

        session_data = self._create_session_data(
            training_data, label_id_dict, attribute=MESSAGE_RESPONSE_ATTRIBUTE
        )

        self.check_input_dimension_consistency(session_data)

        return session_data

    async def process(self, message: "Message", **kwargs: Any) -> None:
        """Return the most likely response and its similarity to the input."""

        label, label_ranking = await self.predict_label(message)

        selector_key = (
            self.retrieval_intent
            if self.retrieval_intent
            else DEFAULT_OPEN_UTTERANCE_TYPE
        )

        logger.debug(
            f"Adding following selector key to message property: {selector_key}"
        )

        prediction_dict = {"response": label, "ranking": label_ranking}

        self._set_message_property(message, prediction_dict, selector_key)



import copy
import logging
import re
import json
from collections import defaultdict

from rasa.core.trackers import DialogueStateTracker
from typing import Text, Any, Dict, Optional, List

from rasa.core.nlg.generator import NaturalLanguageGenerator
from rasa.core.nlg.interpolator import interpolate_text, interpolate
from rasa.utils.timer import timeit

logger = logging.getLogger(__name__)


class TemplatedNaturalLanguageGenerator(NaturalLanguageGenerator):
    """Natural language generator that generates messages based on templates.

    The templates can use variables to customize the utterances based on the
    state of the dialogue."""

    def __init__(self, templates: Dict[Text, List[Dict[Text, Any]]]) -> None:
        self.templates = templates

    def _templates_for_utter_action(self, utter_action, output_channel):
        """Return array of templates that fit the channel and action."""

        channel_templates = []
        default_templates = []

        for template in self.templates[utter_action]:
            if template.get("channel") == output_channel:
                channel_templates.append(template)
            elif not template.get("channel"):
                default_templates.append(template)

        # always prefer channel specific templates over default ones
        if channel_templates:
            return channel_templates
        else:
            return default_templates

    # noinspection PyUnusedLocal
    def _random_template_for(
        self, utter_action: Text, output_channel: Text
    ) -> Optional[Dict[Text, Any]]:
        """Select random template for the utter action from available ones.

        If channel-specific templates for the current output channel are given,
        only choose from channel-specific ones.
        """
        import numpy as np

        if utter_action in self.templates:
            suitable_templates = self._templates_for_utter_action(
                utter_action, output_channel
            )

            if suitable_templates:
                return np.random.choice(suitable_templates)
            else:
                return None
        else:
            return None

    # @timeit
    async def generate(
        self,
        template_name: Text,
        tracker: DialogueStateTracker,
        output_channel: Text,
        **kwargs: Any,
    ) -> Optional[Dict[Text, Any]]:
        """Generate a response for the requested template."""

        filled_slots = tracker.current_slot_values()
        return self.generate_from_slots(
            template_name, filled_slots, output_channel, **kwargs
        )

    def generate_from_slots(
        self,
        template_name: Text,
        filled_slots: Dict[Text, Any],
        output_channel: Text,
        **kwargs: Any,
    ) -> Optional[Dict[Text, Any]]:
        """Generate a response for the requested template."""

        # Fetching a random template for the passed template name
        r = copy.deepcopy(self._random_template_for(template_name, output_channel))
        # Filling the slots in the template and returning the template
        if r is not None:
            return self._fill_template(r, filled_slots, **kwargs)
        else:
            return None

    def _fill_template(
        self,
        template: Dict[Text, Any],
        filled_slots: Optional[Dict[Text, Any]] = None,
        **kwargs: Any,
    ) -> Dict[Text, Any]:
        """"Combine slot values and key word arguments to fill templates."""

        # Getting the slot values in the template variables
        template_vars = self._template_variables(filled_slots, kwargs)

        keys_to_interpolate = [
            "text",
            "image",
            "custom",
            "button",
            "attachment",
            "quick_replies",
        ]
        # Filling the template variables in the template text
        if template_vars and "text" in template:
            try:
                # transforming template tags from
                # "{tag_name}" to "{0[tag_name]}"
                # as described here:
                # https://stackoverflow.com/questions/7934620/python-dots-in-the-name-of-variable-in-a-format-string#comment9695339_7934969
                # assuming that slot_name do not contain newline character here
                text = re.sub(r"{([^\n]+?)}", r"{0[\1]}", template["text"])
                template["text"] = text.format(template_vars)
            except KeyError as e:
                logger.exception(
                    "Failed to fill utterance template '{}'. "
                    "Tried to replace '{}' but could not find "
                    "a value for it. There is no slot with this "
                    "name nor did you pass the value explicitly "
                    "when calling the template. Return template "
                    "without filling the template. "
                    "".format(template, e.args[0])
                )
        if template_vars and "custom" in template:
            try:
                self.iterateNestedDict(template["custom"], template_vars)
                # transforming template tags from
                # "{tag_name}" to "{0[tag_name]}"
                # as described here:
                # https://stackoverflow.com/questions/7934620/python-dots-in-the-name-of-variable-in-a-format-string#comment9695339_7934969
                # assuming that slot_name do not contain newline character here
            except KeyError as e:
                logger.exception(
                    "Failed to fill utterance template '{}'. "
                    "Tried to replace '{}' but could not find "
                    "a value for it. There is no slot with this "
                    "name nor did you pass the value explicitly "
                    "when calling the template. Return template "
                    "without filling the template. "
                    "".format(template, e.args[0])
                )
        if template_vars:
            for key in keys_to_interpolate:
                if key in template:
                    template[key] = interpolate(template[key], template_vars)

        # Inject the language attribute into the template based on the current session_language
        if template.get('custom'):
            template['custom']['language'] = template_vars.get('session_language') or 'English'
            
        return template

    def iterateNestedDict(self, custom: Dict[Text, Any], template_vars: Dict[Text, Any]):
        for k, v in custom.items():
            if isinstance(v, dict):
                self.iterateNestedDict(v, template_vars)
            elif isinstance(v, list):
                self.iterateNestedList(v, template_vars)
            else:
                if k in ['text', 'isVisible', 'file']:
                    v = re.sub(r"{([^\n]+?)}", r"{0[\1]}", v)
                    v = v.format(template_vars)
                    custom[k] = v

    def iterateNestedList(self, custom: List, template_vars: Dict[Text, Any]):
        for i in custom:
            if isinstance(i, list):
                self.iterateNestedList(i, template_vars)
            elif isinstance(i, dict):
                self.iterateNestedDict(i, template_vars)

    @staticmethod
    def _template_variables(
        filled_slots: Dict[Text, Any], kwargs: Dict[Text, Any]
    ) -> Dict[Text, Any]:
        """Combine slot values and key word arguments to fill templates."""

        if filled_slots is None:
            filled_slots = {}

        # Copying the filled slots in the template variables.
        template_vars = filled_slots.copy()
        template_vars.update(kwargs)
        return template_vars



"""
Script used to publish GitHub release notes extracted from CHANGELOG.rst.
This script is executed by Travis after a new release was successfully built.

Uses the following environment variables:
* TRAVIS_TAG: the name of the tag of the current commit.
* GITHUB_TOKEN: a personal access token with 'repo' permissions.

The script also requires ``pandoc`` to be previously installed in the system.
Requires Python3.6+.

Based on code from pytest.
https://github.com/pytest-dev/pytest/blob/master/scripts/publish_gh_release_notes.py
Copyright Holger Krekel and others, 2004-2019.

Distributed under the terms of the MIT license, pytest is free and open source software.
"""
import os
import re
import sys
from pathlib import Path
from typing import Text

# if this needs any more dependencies, they need to be installed on travis deploy stage
import github3
import pypandoc


def create_github_release(slug: Text, token: Text, tag_name: Text, body: Text):
    """Create a github release."""

    github = github3.login(token=token)
    owner, repo = slug.split("/")
    repo = github.repository(owner, repo)
    return repo.create_release(tag_name=tag_name, body=body)


def parse_changelog(tag_name: Text) -> Text:
    """Read the changelog and extract the most recently release entry."""

    p = Path(__file__).parent.parent / "CHANGELOG.rst"
    changelog_lines = p.read_text(encoding="UTF-8").splitlines()

    title_regex = re.compile(r"\[\d+\.\d+\.\d+(\S*)\]\s*-\s*\d{4}-\d{2}-\d{2}")
    consuming_version = False
    version_lines = []
    for line in changelog_lines:
        m = title_regex.match(line)
        if m:
            # found the version we want: start to consume lines
            # until we find the next version title
            if m.group(1) == tag_name:
                consuming_version = True
            # found a new version title while parsing the version we want: break out
            elif consuming_version:
                break
        if consuming_version:
            version_lines.append(line)

    return "\n".join(version_lines)


def convert_rst_to_md(text):
    return pypandoc.convert_text(text, "md", format="rst")


def main():
    tag_name = os.environ.get("TRAVIS_TAG")
    if not tag_name:
        print("environment variable TRAVIS_TAG not set", file=sys.stderr)
        return 1

    token = os.environ.get("GITHUB_TOKEN")
    if not token:
        print("GITHUB_TOKEN not set", file=sys.stderr)
        return 1

    slug = os.environ.get("TRAVIS_REPO_SLUG")
    if not slug:
        print("TRAVIS_REPO_SLUG not set", file=sys.stderr)
        return 1

    rst_body = parse_changelog(tag_name)
    md_body = convert_rst_to_md(rst_body)
    if not create_github_release(slug, token, tag_name, md_body):
        print("Could not publish release notes:", file=sys.stderr)
        print(md_body, file=sys.stderr)
        return 5

    print()
    print(f"Release notes for {tag_name} published successfully:")
    print(f"https://github.com/{slug}/releases/tag/{tag_name}")
    print()
    return 0


if __name__ == "__main__":
    sys.exit(main())



from pygments.lexer import RegexLexer, bygroups, using, default, include
from pygments.lexers.data import JsonLexer
from pygments.token import Keyword, Comment, Token, Text, Generic, Name


class StoryLexer(RegexLexer):
    """Lexer for the Rasa Core story file format.
    Used for syntax highlighting of story snippets in the docs."""

    name = "Story"
    aliases = ["story"]
    filenames = ["*.md"]

    tokens = {
        "comment": [
            (
                r"(\s*<!--)((?:.*?\n?)*)(-->)",
                bygroups(Keyword, Comment.MultiLine, Keyword),
            )
        ],
        "root": [
            include("comment"),
            (r"\s*-\s*(slot)", Token.Operator, ("event", "event_rx")),
            (r"\s*-\s*(restart)", Token.Operator, ("event", "event_rx")),
            (r"\s*-\s*(rewind)", Token.Operator, ("event", "event_rx")),
            (r"\s*-\s*(reset_slots)", Token.Operator, ("event", "event_rx")),
            (r"\s*-\s*(reminder)", Token.Operator, ("event", "event_rx")),
            (r"\s*-\s*(undo)", Token.Operator, ("event", "event_rx")),
            (r"\s*-\s*(export)", Token.Operator, ("event", "event_rx")),
            (r"\s*-\s*(pause)", Token.Operator, ("event", "event_rx")),
            (r"\s*-\s*(resume)", Token.Operator, ("event", "event_rx")),
            (r"\s*-\s*(utter_[^\s]*)", Token.Text, ("event", "event_rx")),
            (
                r"(\s*-(?:\s*)(?:.*?))(\s*)(?:(?:(<!--)" r"((?:.*?\n?)*)(-->))|(\n|$))",
                bygroups(Text, Text, Keyword, Comment.MultiLine, Keyword, Text),
            ),
            (r"\s*\>\s*[^\s]*", Name.Constant),
            (
                r"(#+(?:\s*)(?:.*?))(\s*)(?:(?:(<!--)((?:.*?\n?)*)(-->))|(\n|$))",
                bygroups(
                    Generic.Heading, Text, Keyword, Comment.MultiLine, Keyword, Text
                ),
            ),
            (r"\s*\*\s*", Name.Variable.Magic, ("intent", "intent_rx")),
            (r".*\n", Text),
        ],
        "event": [include("comment"), (r"\s*(\n|$)", Text, "#pop")],
        "event_rx": [(r"({.*?})?", bygroups(using(JsonLexer)), "#pop")],
        "intent": [
            (r"\s*OR\s*", Keyword, "intent_rx"),
            include("comment"),
            (r"\s*(?:\n|$)", Text, "#pop"),
            default("#pop"),
        ],
        "intent_rx": [
            (r'["\'].*["\']', Name.Variable.Magic, "#pop"),
            (
                r"([^\s\{]*\s*)({.*?})?",
                bygroups(Name.Variable.Magic, using(JsonLexer)),
                "#pop",
            ),
            (r"\s*(\n|$)", Text, "#pop:2"),
        ],
    }



"""Prepare a Rasa OSS release.

- creates a release branch
- creates a new changelog section in CHANGELOG.rst based on all collected changes
- increases the version number
- pushes the new branch to GitHub
"""
import argparse
import os
import sys
from pathlib import Path
from subprocess import CalledProcessError, check_call, check_output
from typing import Text, Set

import questionary
import semantic_version
from semantic_version import Version

VERSION_FILE_PATH = "rasa/version.py"

REPO_BASE_URL = "https://github.com/RasaHQ/rasa"

RELEASE_BRANCH_PREFIX = "prepare-release-"


def create_argument_parser() -> argparse.ArgumentParser:
    """Parse all the command line arguments for the release script."""

    parser = argparse.ArgumentParser(description="prepare the next library release")
    parser.add_argument(
        "--next_version",
        type=str,
        help="Either next version number or 'major', 'minor', 'patch'",
    )

    return parser


def project_root() -> Path:
    """Root directory of the project."""
    return Path(os.path.dirname(__file__)).parent


def version_file_path() -> Path:
    """Path to the python file containing the version number."""
    return project_root() / VERSION_FILE_PATH


def write_version_file(version: Text) -> None:
    """Dump a new version into the python version file."""

    with version_file_path().open("w") as f:
        f.write(
            f"# this file will automatically be changed,\n"
            f"# do not add anything but the version number here!\n"
            f'__version__ = "{version}"\n'
        )
    check_call(["git", "add", str(version_file_path().absolute())])


def get_current_version() -> Text:
    """Return the current library version."""

    if not version_file_path().is_file():
        raise FileNotFoundError(
            f"Failed to find version file at {version_file_path().absolute()}"
        )

    # context in which we evaluate the version py -
    # to be able to access the defined version, it already needs to live in the
    # context passed to exec
    _globals = {"__version__": ""}
    with version_file_path().open() as f:
        exec(f.read(), _globals)

    return _globals["__version__"]


def confirm_version(version: Text) -> bool:
    """Allow the user to confirm the version number."""

    if version in git_existing_tags():
        confirmed = questionary.confirm(
            f"Tag with version '{version}' already exists, overwrite?", default=False
        ).ask()
    else:
        confirmed = questionary.confirm(
            f"Current version is '{get_current_version()}. "
            f"Is the next version '{version}' correct ?",
            default=True,
        ).ask()
    if confirmed:
        return True
    else:
        print("Aborting.")
        sys.exit(1)


def ask_version() -> Text:
    """Allow the user to confirm the version number."""

    def is_valid_version_number(v: Text) -> bool:
        # noinspection PyBroadException
        try:
            return v in {"major", "minor", "patch"} or Version.coerce(v) is not None
        except Exception:
            # "coerce" did fail, this is probably not a valid version number
            return False

    version = questionary.text(
        "What is the version number you want to release "
        "('major', 'minor', 'patch' or valid version number]?",
        validate=is_valid_version_number,
    ).ask()

    if version:
        return version
    else:
        print("Aborting.")
        sys.exit(1)


def git_existing_tags() -> Set[Text]:
    """Return all existing tags in the local git repo."""

    stdout = check_output(["git", "tag"])
    return set(stdout.decode().split("\n"))


def git_current_branch() -> Text:
    """Returns the current git branch of the local repo."""

    try:
        output = check_output(["git", "symbolic-ref", "--short", "HEAD"])
        return output.decode().strip()
    except CalledProcessError:
        # e.g. we are in detached head state
        return "master"


def create_release_branch(version: Text) -> Text:
    """Create a new branch for this release. Returns the branch name."""

    branch = f"{RELEASE_BRANCH_PREFIX}{version}"
    check_call(["git", "checkout", "-b", branch])
    return branch


def create_commit(version: Text) -> None:
    """Creates a git commit with all stashed changes."""
    check_call(["git", "commit", "-m", f"prepared release of version {version}"])


def push_changes() -> None:
    """Pushes the current branch to origin."""
    check_call(["git", "push", "origin", "HEAD"])


def ensure_clean_git() -> None:
    """Makes sure the current working git copy is clean."""

    try:
        check_call(["git", "diff-index", "--quiet", "HEAD", "--"])
    except CalledProcessError:
        print("Your git is not clean. Release script can only be run from a clean git.")
        sys.exit(1)


def parse_next_version(version: Text) -> Text:
    """Find the next version as a proper semantic version string."""
    if version == "major":
        return str(Version.coerce(get_current_version()).next_major())
    elif version == "minor":
        return str(Version.coerce(get_current_version()).next_minor())
    elif version == "patch":
        return str(Version.coerce(get_current_version()).next_patch())
    elif semantic_version.validate(version):
        return version
    else:
        raise Exception(f"Invalid version number '{cmdline_args.next_version}'.")


def next_version(args: argparse.Namespace) -> Text:
    """Take cmdline args or ask the user for the next version and return semver."""
    return parse_next_version(args.next_version or ask_version())


def generate_changelog(version: Text) -> None:
    """Call tonwcrier and create a changelog from all available changelog entries."""
    check_call(["towncrier", "--yes", "--version", version], cwd=str(project_root()))


def print_done_message(branch: Text, base: Text, version: Text) -> None:
    """Print final information for the user on what to do next."""

    pull_request_url = f"{REPO_BASE_URL}/compare/{base}...{branch}?expand=1"

    print()
    print(f"\033[94m All done - changes for version {version} are ready! \033[0m")
    print()
    print(f"Please open a PR on GitHub: {pull_request_url}")


def main(args: argparse.Namespace) -> None:
    """Start a release preparation."""

    print(
        "The release script will increase the version number, "
        "create a changelog and create a release branch. Let's go!"
    )

    ensure_clean_git()
    version = next_version(args)
    confirm_version(version)

    write_version_file(version)

    generate_changelog(version)
    base = git_current_branch()
    branch = create_release_branch(version)

    create_commit(version)
    push_changes()

    print_done_message(branch, base, version)


if __name__ == "__main__":
    arg_parser = create_argument_parser()
    cmdline_args = arg_parser.parse_args()
    main(cmdline_args)



DEFAULT_SERVER_PORT = 5005

DEFAULT_SERVER_FORMAT = "{}://localhost:{}"

DEFAULT_SERVER_URL = DEFAULT_SERVER_FORMAT.format("http", DEFAULT_SERVER_PORT)

DEFAULT_NLU_FALLBACK_THRESHOLD = 0.0

DEFAULT_CORE_FALLBACK_THRESHOLD = 0.0

DEFAULT_FALLBACK_ACTION = "action_default_fallback"

DEFAULT_REQUEST_TIMEOUT = 60 * 5  # 5 minutes

DEFAULT_LOCK_LIFETIME = 60  # in seconds

REQUESTED_SLOT = "requested_slot"

# slots for knowledge base
SLOT_LISTED_ITEMS = "knowledge_base_listed_objects"
SLOT_LAST_OBJECT = "knowledge_base_last_object"
SLOT_LAST_OBJECT_TYPE = "knowledge_base_last_object_type"
DEFAULT_KNOWLEDGE_BASE_ACTION = "action_query_knowledge_base"

# start of special user message section
INTENT_MESSAGE_PREFIX = "/"

USER_INTENT_RESTART = "restart"
USER_INTENT_DISCONNECT = "disconnect"

USER_INTENT_BACK = "back"

USER_INTENT_OUT_OF_SCOPE = "out_of_scope"

ACTION_NAME_SENDER_ID_CONNECTOR_STR = "__sender_id:"

BEARER_TOKEN_PREFIX = "Bearer "

# the lowest priority intended to be used by machine learning policies
DEFAULT_POLICY_PRIORITY = 1
# the priority intended to be used by mapping policies
MAPPING_POLICY_PRIORITY = 2
# the priority intended to be used by memoization policies
# it is higher than default and mapping to prioritize training stories
MEMOIZATION_POLICY_PRIORITY = 3
# the priority intended to be used by fallback policies
# it is higher than memoization to prioritize fallback
FALLBACK_POLICY_PRIORITY = 4
# the priority intended to be used by form policies
# it is the highest to prioritize form to the rest of the policies
FORM_POLICY_PRIORITY = 5
UTTER_PREFIX = "utter_"
RESPOND_PREFIX = "respond_"



import logging
from typing import Any, Dict, Text, Optional

from rasa.utils.endpoints import EndpointConfig

logger = logging.getLogger(__name__)


class EventChannel:
    @classmethod
    def from_endpoint_config(cls, broker_config: EndpointConfig) -> "EventChannel":
        raise NotImplementedError(
            "Event broker must implement the `from_endpoint_config` method."
        )

    def flush(self):
        pass
    
    def publish(self, event: Dict[Text, Any]) -> None:
        """Publishes a json-formatted Rasa Core event into an event queue."""

        raise NotImplementedError("Event broker must implement the `publish` method.")



import io
import os
import pickle
import pytest
import tempfile
import shutil
import rasa.utils.io as io_utils
from rasa.nlu import utils


@pytest.fixture(scope="function")
def empty_model_dir():
    temp_path = tempfile.mkdtemp()
    yield temp_path
    if os.path.exists(temp_path):
        shutil.rmtree(temp_path)


@pytest.fixture
def fake_model_dir(empty_model_dir):
    metadata_file = "metadata.json"
    metadata_content = {"pipeline": "pretrained_embeddings_spacy", "language": "en"}
    metadata_path = os.path.join(empty_model_dir, metadata_file)
    utils.write_json_to_file(metadata_path, metadata_content)

    fake_obj = {"Fake", "model"}
    fake_obj_path = os.path.join(empty_model_dir, "component.pkl")
    with open(fake_obj_path, "wb") as f:
        pickle.dump(fake_obj, f)
    return empty_model_dir  # not empty anymore ;)


def test_relative_normpath():
    test_file = "/my/test/path/file.txt"
    assert utils.relative_normpath(test_file, "/my/test") == "path/file.txt"
    assert utils.relative_normpath(None, "/my/test") is None


def test_list_files_invalid_resource():
    with pytest.raises(ValueError) as execinfo:
        io_utils.list_files(None)
    assert "must be a string type" in str(execinfo.value)


def test_list_files_non_existing_dir():
    with pytest.raises(ValueError) as execinfo:
        io_utils.list_files("my/made_up/path")
    assert "Could not locate the resource" in str(execinfo.value)


def test_list_files_ignores_hidden_files(tmpdir):
    # create a hidden file
    open(os.path.join(tmpdir.strpath, ".hidden"), "a").close()
    # create a normal file
    normal_file = os.path.join(tmpdir.strpath, "normal_file")
    open(normal_file, "a").close()
    assert io_utils.list_files(tmpdir.strpath) == [normal_file]


def test_creation_of_existing_dir(tmpdir):
    # makes sure there is no exception
    assert io_utils.create_directory(tmpdir.strpath) is None


def test_ordered():
    target = {"a": [1, 3, 2], "c": "a", "b": 1}
    assert utils.ordered(target) == [("a", [1, 2, 3]), ("b", 1), ("c", "a")]


def test_empty_is_model_dir(empty_model_dir):
    assert utils.is_model_dir(empty_model_dir)


def test_non_existent_folder_is_no_model_dir():
    assert not utils.is_model_dir("nonexistent_for_sure_123/")


def test_data_folder_is_no_model_dir():
    assert not utils.is_model_dir("data/")


def test_model_folder_is_model_dir(fake_model_dir):
    assert utils.is_model_dir(fake_model_dir)


def test_remove_model_empty(empty_model_dir):
    assert utils.remove_model(empty_model_dir)


def test_remove_model_with_files(fake_model_dir):
    assert utils.remove_model(fake_model_dir)


def test_remove_model_invalid(empty_model_dir):
    test_file = "something.else"
    test_content = "Some other stuff"
    test_file_path = os.path.join(empty_model_dir, test_file)
    utils.write_to_file(test_file_path, test_content)

    with pytest.raises(ValueError):
        utils.remove_model(empty_model_dir)

    os.remove(test_file_path)


def test_is_url():
    assert not utils.is_url("./some/file/path")
    assert utils.is_url("https://rasa.com/")



import json
import logging
import os

import typing
from typing import Optional, Text

from rasa.nlu import utils
from rasa.nlu.training_data.formats import markdown
from rasa.nlu.training_data.formats.dialogflow import (
    DIALOGFLOW_AGENT,
    DIALOGFLOW_ENTITIES,
    DIALOGFLOW_ENTITY_ENTRIES,
    DIALOGFLOW_INTENT,
    DIALOGFLOW_INTENT_EXAMPLES,
    DIALOGFLOW_PACKAGE,
)
from rasa.utils.endpoints import EndpointConfig
import rasa.utils.io as io_utils
import re

if typing.TYPE_CHECKING:
    from rasa.nlu.training_data import TrainingData
    from rasa.nlu.training_data.formats.readerwriter import TrainingDataReader

logger = logging.getLogger(__name__)

# Different supported file formats and their identifier
WIT = "wit"
LUIS = "luis"
RASA = "rasa_nlu"
MARKDOWN = "md"
UNK = "unk"
MARKDOWN_NLG = "nlg.md"
DIALOGFLOW_RELEVANT = {DIALOGFLOW_ENTITIES, DIALOGFLOW_INTENT}

_markdown_section_markers = [f"## {s}:" for s in markdown.available_sections]
_json_format_heuristics = {
    WIT: lambda js, fn: "data" in js and isinstance(js.get("data"), list),
    LUIS: lambda js, fn: "luis_schema_version" in js,
    RASA: lambda js, fn: "rasa_nlu_data" in js,
    DIALOGFLOW_AGENT: lambda js, fn: "supportedLanguages" in js,
    DIALOGFLOW_PACKAGE: lambda js, fn: "version" in js and len(js) == 1,
    DIALOGFLOW_INTENT: lambda js, fn: "responses" in js,
    DIALOGFLOW_ENTITIES: lambda js, fn: "isEnum" in js,
    DIALOGFLOW_INTENT_EXAMPLES: lambda js, fn: "_usersays_" in fn,
    DIALOGFLOW_ENTITY_ENTRIES: lambda js, fn: "_entries_" in fn,
}

# looks for pattern like:
# ##
# * intent/response_key
#   - response_text
_nlg_markdown_marker_regex = re.compile(r"##\s*.*\n\*[^:]*\/.*\n\s*\t*\-.*")


def load_data(resource_name: Text, language: Optional[Text] = "en") -> "TrainingData":
    """Load training data from disk.

    Merges them if loaded from disk and multiple files are found."""
    from rasa.nlu.training_data import TrainingData

    if not os.path.exists(resource_name):
        raise ValueError(f"File '{resource_name}' does not exist.")

    files = io_utils.list_files(resource_name)
    data_sets = [_load(f, language) for f in files]
    data_sets = [ds for ds in data_sets if ds]
    if len(data_sets) == 0:
        training_data = TrainingData()
    elif len(data_sets) == 1:
        training_data = data_sets[0]
    else:
        training_data = data_sets[0].merge(*data_sets[1:])

    if training_data.nlg_stories:
        training_data.fill_response_phrases()

    return training_data


async def load_data_from_endpoint(
    data_endpoint: EndpointConfig, language: Optional[Text] = "en"
) -> "TrainingData":
    """Load training data from a URL."""
    import requests

    if not utils.is_url(data_endpoint.url):
        raise requests.exceptions.InvalidURL(data_endpoint.url)
    try:
        response = await data_endpoint.request("get")
        response.raise_for_status()
        temp_data_file = io_utils.create_temporary_file(response.content, mode="w+b")
        training_data = _load(temp_data_file, language)

        return training_data
    except Exception as e:
        logger.warning(f"Could not retrieve training data from URL:\n{e}")


def _reader_factory(fformat: Text) -> Optional["TrainingDataReader"]:
    """Generates the appropriate reader class based on the file format."""
    from rasa.nlu.training_data.formats import (
        MarkdownReader,
        WitReader,
        LuisReader,
        RasaReader,
        DialogflowReader,
        NLGMarkdownReader,
    )

    reader = None
    if fformat == LUIS:
        reader = LuisReader()
    elif fformat == WIT:
        reader = WitReader()
    elif fformat in DIALOGFLOW_RELEVANT:
        reader = DialogflowReader()
    elif fformat == RASA:
        reader = RasaReader()
    elif fformat == MARKDOWN:
        reader = MarkdownReader()
    elif fformat == MARKDOWN_NLG:
        reader = NLGMarkdownReader()
    return reader


def _load(filename: Text, language: Optional[Text] = "en") -> Optional["TrainingData"]:
    """Loads a single training data file from disk."""

    fformat = guess_format(filename)
    if fformat == UNK:
        raise ValueError(f"Unknown data format for file '{filename}'.")

    reader = _reader_factory(fformat)

    if reader:
        return reader.read(filename, language=language, fformat=fformat)
    else:
        return None


def _is_nlg_story_format(content: Text) -> bool:

    match = re.search(_nlg_markdown_marker_regex, content)
    if match:
        return True


def guess_format(filename: Text) -> Text:
    """Applies heuristics to guess the data format of a file.

    Args:
        filename: file whose type should be guessed

    Returns:
        Guessed file format.
    """
    guess = UNK

    content = ""
    try:
        content = io_utils.read_file(filename)
        js = json.loads(content)
    except ValueError:
        if any([marker in content for marker in _markdown_section_markers]):
            guess = MARKDOWN
        elif _is_nlg_story_format(content):
            guess = MARKDOWN_NLG
    else:
        for fformat, format_heuristic in _json_format_heuristics.items():
            if format_heuristic(js, filename):
                guess = fformat
                break

    logger.debug(f"Training data format of '{filename}' is '{guess}'.")

    return guess


def _guess_format(filename: Text) -> Text:
    logger.warning(
        "Using '_guess_format()' is deprecated since Rasa 1.1.5. "
        "Please use 'guess_format()' instead."
    )
    return guess_format(filename)
